import ecs100.*;
import java.awt.Color;
import java.util.*;

public class CityFinder{

    private int listSize = 1000;
    private int distance = 50;

    /**
     * Returns a new list containing every city in targets
     * that is less than dist from some city in sources.
     * The list should not contain any duplicates.
     * Very inefficient
     * - Removes duplicates afterwards.
     * - Answer list gets very big.
     */
    public List<City> findAllCloseCities1(List<City> sources, List<City> targets, double dist){
        List<City> answer = new ArrayList<City>();
        for (City source : sources){
            for (City target : targets){
                if (source.distanceTo(target) < dist){
                    answer.add(target);
                }
            }
        }
        for (int i=0; i<answer.size(); i++){
            // if (i%100==0){UI.printMessage(answer.size());}
            for (int j=i+1; j<answer.size(); j++){
                if (answer.get(i).equals(answer.get(j))){
                    answer.remove(j);
                    j--;
                }
            }
        }
        return answer;
    }


    /**
     * Returns a new list containing every city in targets
     * that is less than dist from some city in sources.
     * The list should not contain any duplicates.
     * Not very efficient
     * - Checks before adding to ensure no duplicates.
     * - answer list doesn't get so big.
     */
    public List<City> findAllCloseCities2(List<City> sources, List<City> targets, double dist){
        List<City> answer = new ArrayList<City>();
        for (City source : sources){
            for (City target : targets){
                if (source.distanceTo(target) < dist && ! answer.contains(target) ){
                    answer.add(target);
                }
            }
        }
        return answer;
    }

    /**
     * Returns a new list containing every city in targets
     * that is less than dist from some city in sources.
     * The list should not contain any duplicates.
     * Very efficient:
     *  - the outer loop checks each target
     *  - breaks inner loop as soon as target is added
     *  - never gets any duplicates => doesn't even have to check
     */
    public List<City> findAllCloseCities3(List<City> sources, List<City> targets, double dist){
        List<City> answer = new ArrayList<City>();
        for (City target : targets){
            for (City source : sources){
                if (target.distanceTo(source) < dist){
                    answer.add(target);
                    break;   // go to the next target and don't look at this one again.
                }
            }
        }
        return answer;
    }



    // GUI and methods to respond.============================
    
    /** Main */
    public static void main(String[] args){
        new CityFinder().setupGUI();
    }
    /** Setup the GUI */
    public void setupGUI(){
        UI.addButton("clear", UI::clearPanes);
        UI.addTextField("Size of lists", (String v)->{this.listSize = Integer.parseInt(v);});
        //UI.addSlider("Distance", 1,250, this.distance,
        //           (double v)->{this.distance = (int)v;});
        UI.addButton("Remove Duplicates", ()->{doFindAllClose(1);});
        UI.addButton("Check B4 adding", ()->{doFindAllClose(2);});
        UI.addButton("Avoid duplicates", ()->{doFindAllClose(3);});
        UI.addButton("Quit", UI::quit);
        UI.setWindowSize(1100, 570);
        UI.setDivider(0.5);
    }

    public void doFindAllClose(int n){
        List<City> sources = this.makeCities(listSize, "s");
        List<City> targets = this.makeCities(listSize, "t");
        UI.clearGraphics();
        double sz = (this.listSize>=10000)?1:2;
        if (listSize <= 100000){
            drawCities(sources, Color.blue, sz);
            drawCities(targets, Color.red, sz);
        }
        Collection<City> close = null;
        //    double distance = Math.max(1,300/Math.sqrt(listSize));
        UI.printf("FindAllClose version %d\n %,d cities; dist<%d ..... ",
                  n, sources.size(), distance);
        javax.swing.JOptionPane.showMessageDialog(UI.getFrame(), "GO?");
        long start = System.currentTimeMillis();
        switch (n) {
        case 1 -> {close = findAllCloseCities1(sources, targets, distance);}
        case 2 -> {close = findAllCloseCities2(sources, targets, distance);}
        case 3 -> {close = findAllCloseCities3(sources, targets, distance);}
        }
        UI.printf("\n    Time: %.3fs\n\n", (System.currentTimeMillis()-start)/1000.0);
        drawCities(close, Color.black, sz);
    }




    public List<City> makeCities(int size, String prefix){
        List<City> ans = new ArrayList<City>();
        for (int i=0; i<size; i++){
            ans.add(makeRandomCity(prefix+i));
        }
        return ans;   
    }

    public City makeRandomCity(String name){
        return new City(name, Math.random()*500+20, Math.random()*500+20);
    }

    /** Draw all the cities in the Collection in the specified color */
    public void drawCities(Collection<City> cities, Color col, double size){
        UI.setColor(col);
        for(City city : cities){
            city.draw(size);
        }
    }


    // UNUSED=====================================

/*
    public void doRemoveFurthest (int n){
        List<City> cities = this.makeCities(listSize, "c");
        City newPlace = this.makeRandomCity("place");
        UI.clearGraphics();
        if (listSize <= 100000){
            drawCities(cities, Color.red, 2);
        }
        UI.setColor(Color.blue);
        newPlace.draw(4);
        long start = System.currentTimeMillis();
        City furthest = null;
        switch (n) {
        case 1 -> {furthest = removeFurthestCity1(cities, newPlace);}
        case 2 -> {furthest = removeFurthestCity2(cities, newPlace);}
        case 3 -> {furthest = removeFurthestCity3(cities, newPlace);}
        }
        UI.printf("Remove furthest (version %d): \n %,d cities: Time: %,d\n\n",
                  n, (cities.size()+1), (System.currentTimeMillis()-start));
        UI.setColor(Color.black);
        furthest.draw(4); 
    }

    /**
     * Remove (and return) the City in cities that is furthest from newPlace
     * This needs to modify the list in cities.
     * Version 1: Recomputes both distances every time.
     
    public City removeFurthestCity1(List<City> cities, City newPlace){
        City furthestCity = cities.get(0);
        for (City city : cities){
            if (newPlace.distanceTo(city) > newPlace.distanceTo(furthestCity)){
                furthestCity = city;
            }
        }
        cities.remove(furthestCity);
        return furthestCity;
    }

    /**
     * Remove (and return) the City in cities that is furthest from newPlace
     * This needs to modify the list in cities.
     * Version 2: more efficient - doesn't recalculate distance to furthest city
     
    public City removeFurthestCity2(List<City> cities, City newPlace){
        City furthestCity = null;
        double maxDist = -1; 
        for (City city : cities){
            double dist = newPlace.distanceTo(city);
            if (dist > maxDist){
                furthestCity = city;
                maxDist = dist;
            }
        }
        cities.remove(furthestCity);   // works even if cities is empty
        return furthestCity;
    }
    
    /**
     * Remove (and return) the City in cities that is furthest from newPlace
     * This needs to modify the list in cities.
     * Version 3: Most efficient version - doesn't search or shift down to remove
     
    public City removeFurthestCity3(List<City> cities, City newPlace){
        int furthestIndex = -1;
        double maxDist = -1; 
        for (int i=0; i<cities.size(); i++){
            double dist = newPlace.distanceTo(cities.get(i));
            if (dist > maxDist){
                furthestIndex = i;
                maxDist = dist;
            }
        }
        if (furthestIndex>=0){
            return cities.set(furthestIndex, cities.remove(cities.size()-1));
        }
        return null;
    }

    /** Setup the GUI 
    public void setupGUI2(){
        UI.addButton("clear", UI::clearPanes);
        UI.addTextField("Size of lists", (String v)->{this.listSize = Integer.parseInt(v);});
        UI.addButton("removeFurthest1", ()->{doRemoveFurthest(1);});
        UI.addButton("removeFurthest2", ()->{doRemoveFurthest(2);});
        UI.addButton("removeFurthest3", ()->{doRemoveFurthest(3);});
        UI.addSlider("Distance", 1,250, this.distance,
                     (double v)->{this.distance = (int)v;});
        UI.addButton("findAllClose1", ()->{doFindAllClose(1);});
        UI.addButton("findAllClose2", ()->{doFindAllClose(2);});
        UI.addButton("findAllClose3", ()->{doFindAllClose(3);});
        UI.addButton("findAllClose4", ()->{doFindAllClose(4);});
        UI.addButton("findAllClose5", ()->{doFindAllClose(5);});
        UI.addButton("Quit", UI::quit);
        UI.setWindowSize(1000, 570);
        UI.setDivider(0.5);
    }
    /**
     * Returns a new list containing every city in targets
     * that is less than dist from some city in sources.
     * The list should not contain any duplicates.
     * Still very inefficient
     * - Removes duplicates afterwards.
     * - But removes items from the end of the list
     
    public List<City> findAllCloseCities1Reverse(List<City> sources, List<City> targets, double dist){
        List<City> answer = new ArrayList<City>();
        for (City source : sources){
            for (City target : targets){
                if (source.distanceTo(target) < dist){
                    answer.add(target);
                }
            }
        }
        for (int i=answer.size()-1; i>0; i--){
            // if (i%100==0){UI.printMessage("Removing duplicates from "+answer.size() +" targets");}
            for (int j=answer.size()-1; j>i; j--){
                if (answer.get(i)==answer.get(j)){
                    answer.remove(j);
                }
            }
        }
        return answer;
    }


    /**
     * Returns a set containing every city in targets
     * that is less than dist from some city in sources.
     * The list should not contain any duplicates.
     * More efficient:
     *  - checks to see if target already in answer to avoid duplicates
     *  - uses a Set so that checking is very fast
     
    public Set<City> findAllCloseCities2Set(List<City> sources, List<City> targets, double dist){
        Set<City> answer = new HashSet<City>();
        for (City source : sources){
            for (City target : targets){
                if (source.distanceTo(target) < dist && ! answer.contains(target) ){
                    answer.add(target);
                }
            }
        }
        return answer;
    }
*/

}           
 
